 /**
  * XMMS client lib mainloop / poll() implementation for Mozilla.
  *
  * TODO Don't know if this works as expected (or at all)
  * TODO Proxy OnSocketReady to mainloop
  *
  * xmms client lib does its own I/O (networking), via a FD, and so needs to be notified by Mozilla
  * when there's new data in the incoming queue, and get notified when the outgoing queue is ready to
  * accept (more) data.
  *
  * Mozilla does the poll() in nsSocketTransportService2.cpp, and offers an acceptSocket(),
  * which takes an nsASocketHandler. We implement the latter.
  * nsSocketTransportService expects an NSPR PRFileDesc from the nsASocketHandler,
  * which unfortunately is not the same as the OS FD, and NSPR's convert method is private.
  *
  * So:
  * - OS FD -> PRFD using private/pprio.h PR_ImportFile()
  * - Implement nsASocketHandler
  * - Hook it up via nsISocketTransportService.attachSocket()
  *
  * http://lxr.mozilla.org/seamonkey/source/nsprpub/pr/include/private/pprio.h#79
  * http://lxr.mozilla.org/seamonkey/source/netwerk/base/public/nsASocketHandler.h#45
  * http://lxr.mozilla.org/seamonkey/source/netwerk/base/public/nsISocketTransportService.idl#84
  */

#include "xmmsclient-moz.h"
#include "nsISocketTransportService.h"
#include "nsCOMPtr.h"
#include "nsServiceManagerUtils.h"
#include "nsThreadUtils.h"
#include <iostream>

static void xmmsSocketHandler_need_out_cb (int need_out, void *data)
{
    std::cout << "xmmsSocketHandler_need_out " << need_out << std::endl;
    xmmsSocketHandler* sh = (xmmsSocketHandler*) data;
    if (!sh)
        return;

    // TODO should be done on socket thread, not main thread
    if (need_out)
        sh->mPollFlags |= PR_POLL_WRITE;
    else
        sh->mPollFlags &= ~PR_POLL_WRITE;
    std::cout << "set poll flags" << std::endl;
    // TODO tell to reset polling?
}

NS_IMPL_ISUPPORTS1(xmmsSocketHandler, nsASocketHandler)

xmmsSocketHandler::xmmsSocketHandler(xmmsc_connection_t *c)
  : nsASocketHandler(),
    Xmms::MainloopInterface::MainloopInterface(c),
    mXMMSConn(c)
{
    running_ = true;
    mPollFlags = PR_POLL_READ | PR_POLL_EXCEPT;
    mPollTimeout = 1000;
}

void xmmsSocketHandler::Init()
{
    int osfd = xmmsc_io_fd_get(mXMMSConn);
    PRFileDesc* prfd = PR_ImportFile(osfd);

    nsCOMPtr<nsISocketTransportService> stransserv =
        do_GetService("@mozilla.org/network/socket-transport-service;1");
    stransserv->AttachSocket(prfd, this);

    xmmsc_io_need_out_callback_set(mXMMSConn, xmmsSocketHandler_need_out_cb, this);

    xmmsSocketHandler_need_out_cb(xmmsc_io_want_out(mXMMSConn), this);
    std::cout << "mainloop inited" << std::endl;
}



class SocketEvent : public nsIRunnable
{
    NS_DECL_ISUPPORTS
public:
    SocketEvent(xmmsc_connection_t* c, xmmsSocketHandler* sh)
        : mXMMSConn(c), mSH(sh) {}
    NS_IMETHOD Run() = 0;
protected:
    xmmsc_connection_t* mXMMSConn;
    xmmsSocketHandler* mSH;
};
NS_IMPL_THREADSAFE_ISUPPORTS1(SocketEvent, nsIRunnable)

class ReadEvent : public SocketEvent
{
public:
    ReadEvent(xmmsc_connection_t* c, xmmsSocketHandler* sh)
        : SocketEvent(c, sh) {}
    NS_IMETHOD Run()
    {
        std::cout << "ReadEvent in main thread" << std::endl;
        xmmsc_io_in_handle(mXMMSConn);
        mSH->mPollFlags |= PR_POLL_READ;
        std::cout << "Read done" << std::endl;
        return NS_OK;
    }
};

class WriteEvent : public SocketEvent
{
public:
    WriteEvent(xmmsc_connection_t* c, xmmsSocketHandler* sh)
        : SocketEvent(c, sh) {}
    NS_IMETHOD Run()
    {
        std::cout << "WriteEvent in main thread" << std::endl;
        xmmsc_io_out_handle(mXMMSConn);
        xmmsSocketHandler_need_out_cb(xmmsc_io_want_out(mXMMSConn), mSH); // TODO done automatically?
        std::cout << "Write done" << std::endl;
        return NS_OK;
    }
};

// TODO does not work - first gets not called at all for ~10s, then tousands of times
void xmmsSocketHandler::OnSocketReady (PRFileDesc *fd, PRInt16 outFlags)
{
    std::cout << "OnSocketReady in socket thread" << std::endl;
    if (outFlags & PR_POLL_READ)
    {
        NS_DispatchToMainThread(new ReadEvent(mXMMSConn, this));
        mPollFlags &= ~PR_POLL_READ;
        std::cout << "ReadReady dispatched" << std::endl;
    }
    if (outFlags & PR_POLL_WRITE)
    {
        NS_DispatchToMainThread(new WriteEvent(mXMMSConn, this));
        mPollFlags &= ~PR_POLL_WRITE;
        std::cout << "WriteReady dispatched" << std::endl;
    }
    // TODO is freed nsIRunnable freed by ThreadManager?
}


class DetachedEvent : public SocketEvent
{
public:
    DetachedEvent(xmmsc_connection_t* c, xmmsSocketHandler* sh)
        : SocketEvent(c, sh) {}
    NS_IMETHOD Run()
    {
        std::cout << "OnSocketDetached in main thread" << std::endl;
        xmmsc_io_disconnect(mXMMSConn);
        std::cout << "OnSocketDetached done" << std::endl;
        return NS_OK;
    }
};

void xmmsSocketHandler::OnSocketDetached (PRFileDesc *fd)
{
    std::cout << "OnSocketDetached in socket thread" << std::endl;
    NS_DispatchToMainThread(new DetachedEvent(mXMMSConn, this));
    std::cout << "OnSocketDetached dispatched" << std::endl;
}
